@@ -105,7 +105,9 @@ use futures_util::future::FutureExt;
 use futures_util::select;
 use futures_util::stream::Stream;
 use futures_util::StreamExt;
+use tls::TlsOptions;
 
+use std::cmp;
 use std::collections::HashMap;
 use std::iter;
 use std::net::{SocketAddr, ToSocketAddrs};
@@ -113,21 +115,20 @@ use std::option;
 use std::pin::Pin;
 use std::slice;
 use std::str::{self, FromStr};
+use std::sync::atomic::{AtomicU64, Ordering};
 use std::sync::Arc;
 use std::task::{Context, Poll};
-use subslice::SubsliceExt;
+use std::time::Duration;
 use tokio::io::ErrorKind;
-use tokio::io::{AsyncRead, AsyncWriteExt};
-use tokio::io::{AsyncReadExt, AsyncWrite};
+use tokio::time::sleep;
 use url::Url;
 
-use bytes::{Buf, Bytes, BytesMut};
+use bytes::{Bytes, BytesMut};
 use serde::{Deserialize, Serialize};
 use serde_repr::{Deserialize_repr, Serialize_repr};
 use tokio::io;
 use tokio::io::BufWriter;
 use tokio::net::TcpStream;
-use tokio::sync::Mutex;
 use tokio::sync::{mpsc, oneshot};
 use tokio::task;
 
@@ -141,14 +142,20 @@ const LANG: &str = "rust";
 /// must be provided using `Options::tls_client_config`.
 pub use tokio_rustls::rustls;
 
+use connection::Connection;
+pub use header::{HeaderMap, HeaderValue};
+
+mod connection;
 mod options;
+
 pub use options::ConnectOptions;
+pub mod header;
 mod tls;
 
 /// Information sent by the server back to this client
 /// during initial connection, and possibly again later.
 #[allow(unused)]
-#[derive(Debug, Deserialize, Default, Clone)]
+#[derive(Debug, Deserialize, Default, Clone, Eq, PartialEq)]
 pub struct ServerInfo {
     /// The unique identifier of the NATS server.
     #[serde(default)]
@@ -201,17 +208,19 @@ pub struct ServerInfo {
     pub lame_duck_mode: bool,
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, Eq, PartialEq)]
 pub(crate) enum ServerOp {
     Ok,
     Info(Box<ServerInfo>),
     Ping,
     Pong,
+    Error(ServerError),
     Message {
         sid: u64,
         subject: String,
         reply: Option<String>,
         payload: Bytes,
+        headers: Option<HeaderMap>,
     },
 }
 
@@ -221,20 +230,21 @@ pub enum Command {
         subject: String,
         payload: Bytes,
         respond: Option<String>,
+        headers: Option<HeaderMap>,
     },
     Subscribe {
         sid: u64,
         subject: String,
         queue_group: Option<String>,
+        sender: mpsc::Sender<Message>,
     },
     Unsubscribe {
-        uid: u64,
+        sid: u64,
         max: Option<u64>,
     },
     Ping,
-    Pong,
     Flush {
-        result: oneshot::Sender<io::Result<()>>,
+        result: oneshot::Sender<Result<(), io::Error>>,
     },
     TryFlush,
     Connect(ConnectInfo),
@@ -247,6 +257,7 @@ pub enum ClientOp {
         subject: String,
         payload: Bytes,
         respond: Option<String>,
+        headers: Option<HeaderMap>,
     },
     Subscribe {
         sid: u64,
@@ -259,234 +270,83 @@ pub enum ClientOp {
     },
     Ping,
     Pong,
-    Flush {
-        result: oneshot::Sender<io::Result<()>>,
-    },
     TryFlush,
     Connect(ConnectInfo),
 }
 
-/// Supertrait enabling trait object for containing both TLS and non TLS `TcpStream` connection.
-trait AsyncReadWrite: AsyncWrite + AsyncRead + Send + Unpin {}
-
-/// Blanked implementation that applies to both TLS and non-TLS `TcpStream`.
-impl<T> AsyncReadWrite for T where T: AsyncRead + AsyncWrite + Unpin + Send {}
-
-/// A framed connection
-pub(crate) struct Connection {
-    stream: Box<dyn AsyncReadWrite>,
-    buffer: BytesMut,
-}
-
-/// Internal representation of the connection.
-/// Helds connection with NATS Server and communicates with `Client` via channels.
-impl Connection {
-    pub(crate) fn try_read_op(&mut self) -> Result<Option<ServerOp>, io::Error> {
-        if self.buffer.starts_with(b"+OK\r\n") {
-            self.buffer.advance(5);
-            return Ok(Some(ServerOp::Ok));
-        }
-
-        if self.buffer.starts_with(b"PING\r\n") {
-            self.buffer.advance(6);
-
-            return Ok(Some(ServerOp::Ping));
-        }
-
-        if self.buffer.starts_with(b"PONG\r\n") {
-            self.buffer.advance(6);
-
-            return Ok(Some(ServerOp::Pong));
-        }
-
-        if self.buffer.starts_with(b"INFO ") {
-            if let Some(len) = self.buffer.find(b"\r\n") {
-                let line = std::str::from_utf8(&self.buffer[5..len])
-                    .map_err(|err| io::Error::new(io::ErrorKind::InvalidInput, err))?;
-
-                let server_info = serde_json::from_str(line)
-                    .map_err(|err| io::Error::new(io::ErrorKind::InvalidInput, err))?;
-
-                self.buffer.advance(len + 2);
-
-                return Ok(Some(ServerOp::Info(Box::new(server_info))));
-            }
-
-            return Ok(None);
-        }
-
-        if self.buffer.starts_with(b"MSG ") {
-            if let Some(len) = self.buffer.find(b"\r\n") {
-                let line = std::str::from_utf8(&self.buffer[4..len]).unwrap();
-                let args = line.split(' ').filter(|s| !s.is_empty());
-                // TODO(caspervonb) we can drop this alloc
-                let args = args.collect::<Vec<_>>();
-
-                // Parse the operation syntax: MSG <subject> <sid> [reply-to] <#bytes>
-                let (subject, sid, reply_to, payload_len) = match args[..] {
-                    [subject, sid, payload_len] => (subject, sid, None, payload_len),
-                    [subject, sid, reply_to, payload_len] => {
-                        (subject, sid, Some(reply_to), payload_len)
-                    }
-                    _ => {
-                        return Err(io::Error::new(
-                            io::ErrorKind::InvalidInput,
-                            "invalid number of arguments after MSG",
-                        ));
-                    }
-                };
-
-                let sid = u64::from_str(sid)
-                    .map_err(|err| io::Error::new(io::ErrorKind::InvalidInput, err))?;
-
-                // Parse the number of payload bytes.
-                let payload_len = usize::from_str(payload_len)
-                    .map_err(|err| io::Error::new(io::ErrorKind::InvalidInput, err))?;
-
-                // Only advance if there is enough data for the entire operation and payload remaining.
-                if len + payload_len + 4 <= self.buffer.remaining() {
-                    let subject = subject.to_owned();
-                    let reply_to = reply_to.map(String::from);
-
-                    self.buffer.advance(len + 2);
-                    let payload = self.buffer.split_to(payload_len).freeze();
-                    self.buffer.advance(2);
-
-                    return Ok(Some(ServerOp::Message {
-                        sid,
-                        reply: reply_to,
-                        subject,
-                        payload,
-                    }));
-                }
-            }
-
-            return Ok(None);
-        }
-
-        Ok(None)
-    }
-
-    pub(crate) async fn read_op(&mut self) -> Result<Option<ServerOp>, io::Error> {
-        loop {
-            if let Some(op) = self.try_read_op()? {
-                return Ok(Some(op));
-            }
-
-            if 0 == self.stream.read_buf(&mut self.buffer).await? {
-                if self.buffer.is_empty() {
-                    return Ok(None);
-                } else {
-                    return Err(io::Error::new(io::ErrorKind::ConnectionReset, ""));
-                }
-            }
-        }
-    }
-
-    pub(crate) async fn write_op(&mut self, item: ClientOp) -> Result<(), io::Error> {
-        match item {
-            ClientOp::Connect(connect_info) => {
-                let op = format!(
-                    "CONNECT {}\r\n",
-                    serde_json::to_string(&connect_info)
-                        .map_err(|err| io::Error::new(io::ErrorKind::InvalidData, err))?
-                );
-                self.stream.write_all(op.as_bytes()).await?;
-            }
-            ClientOp::Publish {
-                subject,
-                payload,
-                respond,
-            } => {
-                let mut bufi = itoa::Buffer::new();
-                self.stream.write_all(b"PUB ").await?;
-                self.stream.write_all(subject.as_bytes()).await?;
-                self.stream.write_all(b" ").await?;
-                if let Some(respond) = respond {
-                    self.stream.write_all(respond.as_bytes()).await?;
-                    self.stream.write_all(b" ").await?;
-                }
-                self.stream
-                    .write_all(bufi.format(payload.len()).as_bytes())
-                    .await?;
-                self.stream.write_all(b"\r\n").await?;
-                self.stream.write_all(&payload).await?;
-                self.stream.write_all(b"\r\n").await?;
-            }
-
-            ClientOp::Subscribe {
-                sid,
-                subject,
-                queue_group,
-            } => {
-                self.stream.write_all(b"SUB ").await?;
-                self.stream.write_all(subject.as_bytes()).await?;
-                if let Some(queue_group) = queue_group {
-                    self.stream
-                        .write_all(format!(" {}", queue_group).as_bytes())
-                        .await?;
-                }
-                self.stream
-                    .write_all(format!(" {}\r\n", sid).as_bytes())
-                    .await?;
-                self.stream.flush().await?;
-            }
-
-            ClientOp::Unsubscribe { sid: id, max } => {
-                self.stream.write_all(b"UNSUB ").await?;
-                self.stream.write_all(format!("{}", id).as_bytes()).await?;
-                if let Some(max) = max {
-                    self.stream
-                        .write_all(format!(" {}", max).as_bytes())
-                        .await?;
-                }
-                self.stream.write_all(b"\r\n").await?;
-            }
-            ClientOp::Ping => {
-                self.stream.write_all(b"PING\r\n").await?;
-                self.stream.flush().await?;
-            }
-            ClientOp::Pong => {
-                self.stream.write_all(b"PONG\r\n").await?;
-                self.stream.flush().await?;
-            }
-            ClientOp::Flush { result } => {
-                result.send(self.stream.flush().await).map_err(|_| {
-                    io::Error::new(io::ErrorKind::Other, "one shot failed to be received")
-                })?;
-            }
-            ClientOp::TryFlush => {
-                self.stream.flush().await?;
-            }
-        }
-
-        Ok(())
-    }
-}
-
 /// Maintains a list of servers and establishes connections.
 pub(crate) struct Connector {
-    server_addrs: Vec<ServerAddr>,
-    options: ConnectOptions,
+    /// A map of servers and number of connect attempts.
+    servers: HashMap<ServerAddr, usize>,
+    options: TlsOptions,
 }
 
 impl Connector {
+    pub(crate) fn new<A: ToServerAddrs>(
+        addrs: A,
+        options: TlsOptions,
+    ) -> Result<Connector, io::Error> {
+        let servers = addrs
+            .to_server_addrs()?
+            .into_iter()
+            .map(|addr| (addr, 0))
+            .collect();
+
+        Ok(Connector { servers, options })
+    }
+
     pub(crate) async fn connect(&mut self) -> Result<(ServerInfo, Connection), io::Error> {
-        loop {
+        for _ in 0..128 {
             if let Ok(inner) = self.try_connect().await {
                 return Ok(inner);
             }
         }
+
+        Err(io::Error::new(io::ErrorKind::Other, "unable to connect"))
     }
 
     pub(crate) async fn try_connect(&mut self) -> Result<(ServerInfo, Connection), io::Error> {
         let mut error = None;
 
-        for server_addr in &self.server_addrs {
-            match self.try_connect_to(server_addr).await {
-                Ok(inner) => return Ok(inner),
-                Err(inner) => error.replace(inner),
+        let server_addrs: Vec<ServerAddr> = self.servers.keys().cloned().collect();
+        for server_addr in server_addrs {
+            let server_attempts = self.servers.get_mut(&server_addr).unwrap();
+            let duration = if *server_attempts == 0 {
+                Duration::from_millis(0)
+            } else {
+                let exp: u32 = (*server_attempts - 1).try_into().unwrap_or(std::u32::MAX);
+                let max = Duration::from_secs(4);
+
+                cmp::min(Duration::from_millis(2_u64.saturating_pow(exp)), max)
             };
+
+            *server_attempts += 1;
+            sleep(duration).await;
+
+            let socket_addrs = server_addr.socket_addrs()?;
+            for socket_addr in socket_addrs {
+                match self
+                    .try_connect_to(&socket_addr, server_addr.tls_required(), server_addr.host())
+                    .await
+                {
+                    Ok((info, connection)) => {
+                        for url in &info.connect_urls {
+                            let server_addr = url.parse::<ServerAddr>()?;
+
+                            if !self.servers.contains_key(&server_addr) {
+                                self.servers.insert(server_addr, 0);
+                            }
+                        }
+
+                        let server_attempts = self.servers.get_mut(&server_addr).unwrap();
+                        *server_attempts = 0;
+
+                        return Ok((info, connection));
+                    }
+
+                    Err(inner) => error.replace(inner),
+                };
+            }
         }
 
         Err(error.unwrap())
@@ -494,11 +354,13 @@ impl Connector {
 
     pub(crate) async fn try_connect_to(
         &self,
-        server_addr: &ServerAddr,
+        socket_addr: &SocketAddr,
+        tls_required: bool,
+        tls_host: &str,
     ) -> Result<(ServerInfo, Connection), io::Error> {
         let tls_config = tls::config_tls(&self.options).await?;
 
-        let tcp_stream = TcpStream::connect((server_addr.host(), server_addr.port())).await?;
+        let tcp_stream = TcpStream::connect(socket_addr).await?;
         tcp_stream.set_nodelay(true)?;
 
         let mut connection = Connection {
@@ -523,10 +385,7 @@ impl Connector {
             }
         };
 
-        let tls_required =
-            self.options.tls_required || info.tls_required || server_addr.tls_required();
-
-        if tls_required {
+        if self.options.tls_required || info.tls_required || tls_required {
             let tls_config = Arc::new(tls_config);
             let tls_connector =
                 tokio_rustls::TlsConnector::try_from(tls_config).map_err(|err| {
@@ -537,7 +396,7 @@ impl Connector {
                 })?;
 
             let domain = rustls::ServerName::try_from(info.host.as_str())
-                .or_else(|_| rustls::ServerName::try_from(server_addr.host()))
+                .or_else(|_| rustls::ServerName::try_from(tls_host))
                 .map_err(|_| {
                     io::Error::new(
                         ErrorKind::InvalidInput,
@@ -564,69 +423,25 @@ struct Subscription {
     max: Option<u64>,
 }
 
-#[derive(Debug)]
-struct SubscriptionContext {
-    next_sid: u64,
-    next_uid: u64,
-    subscription_map: HashMap<u64, Subscription>,
-    uid_map: HashMap<u64, u64>,
-}
-
-impl SubscriptionContext {
-    fn new() -> SubscriptionContext {
-        SubscriptionContext {
-            next_sid: 1,
-            next_uid: 1,
-            subscription_map: HashMap::new(),
-            uid_map: HashMap::new(),
-        }
-    }
-
-    fn get_mut(&mut self, sid: u64) -> Option<&mut Subscription> {
-        self.subscription_map.get_mut(&sid)
-    }
-
-    fn entry(&mut self, sid: u64) -> std::collections::hash_map::Entry<u64, Subscription> {
-        self.subscription_map.entry(sid)
-    }
-
-    fn insert(&mut self, subscription: Subscription) -> u64 {
-        let sid = self.next_sid;
-        let uid = self.next_uid;
-        self.next_sid += 1;
-        self.next_uid += 1;
-
-        self.subscription_map.insert(sid, subscription);
-        self.uid_map.insert(uid, sid);
-
-        sid
-    }
-    fn remove(&mut self, sid: u64) -> bool {
-        self.subscription_map.remove(&sid).is_some()
-    }
-
-    fn get_sid(&self, uid: u64) -> Option<u64> {
-        self.uid_map.get(&uid).copied()
-    }
-}
-
 /// A connection handler which facilitates communication from channels to a single shared connection.
 pub(crate) struct ConnectionHandler {
     connection: Connection,
     connector: Connector,
-    subscription_context: Arc<Mutex<SubscriptionContext>>,
+    subscriptions: HashMap<u64, Subscription>,
+    events: mpsc::Sender<ServerEvent>,
 }
 
 impl ConnectionHandler {
     pub(crate) fn new(
         connection: Connection,
         connector: Connector,
-        subscription_context: Arc<Mutex<SubscriptionContext>>,
+        events: mpsc::Sender<ServerEvent>,
     ) -> ConnectionHandler {
         ConnectionHandler {
             connection,
             connector,
-            subscription_context,
+            subscriptions: HashMap::new(),
+            events,
         }
     }
 
@@ -653,17 +468,22 @@ impl ConnectionHandler {
                             println!("error handling operation {}", err);
                         }
                         Ok(None) => {
-                            if let Err(err) = self.handle_reconnect().await {
+                            if let Err(err) = self.handle_disconnect().await {
                                 println!("error handling operation {}", err);
+                            } else {
                             }
                         }
-                        Err(_) => {},
+                        Err(err) => {
+                            if let Err(err) = self.handle_disconnect().await {
+                                println!("error handling operation {}", err);
+                            }
+                        },
                     }
                 }
             }
         }
 
-        self.connection.stream.flush().await?;
+        self.connection.flush().await?;
 
         Ok(())
     }
@@ -673,32 +493,32 @@ impl ConnectionHandler {
             ServerOp::Ping => {
                 self.connection.write_op(ClientOp::Pong).await?;
             }
+            ServerOp::Error(error) => {
+                self.events.try_send(ServerEvent::Error(error)).ok();
+            }
             ServerOp::Message {
                 sid,
                 subject,
                 reply,
                 payload,
+                headers,
             } => {
-                let mut context = self.subscription_context.lock().await;
-
-                if let Some(subscription) = context.get_mut(sid) {
+                if let Some(subscription) = self.subscriptions.get_mut(&sid) {
                     let message = Message {
                         subject,
                         reply,
                         payload,
+                        headers,
                     };
 
                     // if the channel for subscription was dropped, remove the
                     // subscription from the map and unsubscribe.
                     if subscription.sender.send(message).await.is_err() {
-                        context.remove(sid);
+                        self.subscriptions.remove(&sid);
                         self.connection
-                            .write_op(ClientOp::Unsubscribe {
-                                sid: sid,
-                                max: None,
-                            })
+                            .write_op(ClientOp::Unsubscribe { sid, max: None })
                             .await?;
-                        self.connection.stream.flush().await?;
+                        self.connection.flush().await?;
                     // if channel was open and we sent the messsage, increase the
                     // `delivered` counter.
                     } else {
@@ -708,7 +528,7 @@ impl ConnectionHandler {
                         // the result, `drop` the `sender` channel.
                         if let Some(max) = subscription.max {
                             if subscription.delivered.ge(&max) {
-                                context.remove(sid);
+                                self.subscriptions.remove(&sid);
                             }
                         }
                     }
@@ -725,49 +545,49 @@ impl ConnectionHandler {
 
     async fn handle_command(&mut self, command: Command) -> Result<(), io::Error> {
         match command {
-            Command::Unsubscribe { uid, max } => {
-                let mut context = self.subscription_context.lock().await;
-                if let Some(sid) = context.get_sid(uid) {
-                    if let Some(subscription) = context.get_mut(sid) {
-                        // check if max was passed to unsubscribe, meaning it is
-                        // unsubscribe_after.
-                        if let Some(max) = max {
-                            // if we already reached unsub_after max delivery limit, remove
-                            // `Subscription`.
-                            if subscription.delivered >= max {
-                                context.remove(sid);
-                            // otherwise just set `max` value for the `Subscription`.
-                            } else {
-                                context.entry(sid).and_modify(|sub| sub.max = Some(max));
+            Command::Unsubscribe { sid, max } => {
+                if let Some(subscription) = self.subscriptions.get_mut(&sid) {
+                    subscription.max = max;
+                    match subscription.max {
+                        Some(n) => {
+                            if subscription.delivered >= n {
+                                self.subscriptions.remove(&sid);
                             }
-                        // if `max` is `None`, just remove the subscription.
-                        } else {
-                            context.remove(sid);
                         }
+                        None => {
+                            self.subscriptions.remove(&sid);
+                        }
+                    }
 
-                        if let Err(err) = self
-                            .connection
-                            .write_op(ClientOp::Unsubscribe { sid, max })
-                            .await
-                        {
-                            println!("Send failed with {:?}", err);
-                        }
+                    if let Err(err) = self
+                        .connection
+                        .write_op(ClientOp::Unsubscribe { sid, max })
+                        .await
+                    {
+                        println!("Send failed with {:?}", err);
                     }
                 }
             }
             Command::Ping => {
-                while let Err(err) = self.connection.write_op(ClientOp::Ping).await {
-                    self.handle_reconnect().await?;
-                }
-            }
-            Command::Pong => {
-                while let Err(err) = self.connection.write_op(ClientOp::Ping).await {
-                    self.handle_reconnect().await?;
+                if let Err(err) = self.connection.write_op(ClientOp::Ping).await {
+                    self.handle_disconnect().await?;
                 }
             }
             Command::Flush { result } => {
-                if let Err(err) = self.connection.write_op(ClientOp::Flush { result }).await {
-                    self.handle_reconnect().await?;
+                if let Err(err) = self.connection.flush().await {
+                    if let Err(err) = self.handle_disconnect().await {
+                        result.send(Err(err)).map_err(|_| {
+                            io::Error::new(io::ErrorKind::Other, "one shot failed to be received")
+                        })?;
+                    } else if let Err(err) = self.connection.flush().await {
+                        result.send(Err(err)).map_err(|_| {
+                            io::Error::new(io::ErrorKind::Other, "one shot failed to be received")
+                        })?;
+                    }
+                } else {
+                    result.send(Ok(())).map_err(|_| {
+                        io::Error::new(io::ErrorKind::Other, "one shot failed to be received")
+                    })?;
                 }
             }
             Command::TryFlush => {
@@ -779,7 +599,18 @@ impl ConnectionHandler {
                 sid,
                 subject,
                 queue_group,
+                sender,
             } => {
+                let subscription = Subscription {
+                    sender,
+                    delivered: 0,
+                    max: None,
+                    subject: subject.to_owned(),
+                    queue_group: queue_group.to_owned(),
+                };
+
+                self.subscriptions.insert(sid, subscription);
+
                 if let Err(err) = self
                     .connection
                     .write_op(ClientOp::Subscribe {
@@ -796,6 +627,7 @@ impl ConnectionHandler {
                 subject,
                 payload,
                 respond,
+                headers,
             } => {
                 while let Err(err) = self
                     .connection
@@ -803,10 +635,11 @@ impl ConnectionHandler {
                         subject: subject.clone(),
                         payload: payload.clone(),
                         respond: respond.clone(),
+                        headers: headers.clone(),
                     })
                     .await
                 {
-                    self.handle_reconnect().await?;
+                    self.handle_disconnect().await?;
                     println!("Sending Publish failed with {:?}", err);
                 }
             }
@@ -816,7 +649,7 @@ impl ConnectionHandler {
                     .write_op(ClientOp::Connect(connect_info.clone()))
                     .await
                 {
-                    self.handle_reconnect().await?;
+                    self.handle_disconnect().await?;
                 }
             }
         }
@@ -824,12 +657,18 @@ impl ConnectionHandler {
         Ok(())
     }
 
+    async fn handle_disconnect(&mut self) -> io::Result<()> {
+        self.events.try_send(ServerEvent::Disconnect).ok();
+        self.handle_reconnect().await?;
+
+        Ok(())
+    }
+
     async fn handle_reconnect(&mut self) -> Result<(), io::Error> {
         let (_, connection) = self.connector.connect().await?;
         self.connection = connection;
 
-        let subscription_context = self.subscription_context.lock().await;
-        for (sid, subscription) in &subscription_context.subscription_map {
+        for (sid, subscription) in &self.subscriptions {
             self.connection
                 .write_op(ClientOp::Subscribe {
                     sid: *sid,
@@ -839,7 +678,7 @@ impl ConnectionHandler {
                 .await
                 .unwrap();
         }
-
+        self.events.try_send(ServerEvent::Reconnect).ok();
         Ok(())
     }
 }
@@ -847,20 +686,17 @@ impl ConnectionHandler {
 /// Client is a `Clonable` handle to NATS connection.
 /// Client should not be created directly. Instead, one of two methods can be used:
 /// [connect] and [ConnectOptions::connect]
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub struct Client {
     sender: mpsc::Sender<Command>,
-    subscription_context: Arc<Mutex<SubscriptionContext>>,
+    next_subscription_id: Arc<AtomicU64>,
 }
 
 impl Client {
-    pub(crate) fn new(
-        sender: mpsc::Sender<Command>,
-        subscription_context: Arc<Mutex<SubscriptionContext>>,
-    ) -> Client {
+    pub(crate) fn new(sender: mpsc::Sender<Command>) -> Client {
         Client {
             sender,
-            subscription_context,
+            next_subscription_id: Arc::new(AtomicU64::new(0)),
         }
     }
 
@@ -870,6 +706,24 @@ impl Client {
                 subject,
                 payload,
                 respond: None,
+                headers: None,
+            })
+            .await?;
+        Ok(())
+    }
+
+    pub async fn publish_with_headers(
+        &mut self,
+        subject: String,
+        headers: HeaderMap,
+        payload: Bytes,
+    ) -> Result<(), Error> {
+        self.sender
+            .send(Command::Publish {
+                subject,
+                payload,
+                respond: None,
+                headers: Some(headers),
             })
             .await?;
         Ok(())
@@ -886,6 +740,25 @@ impl Client {
                 subject,
                 payload,
                 respond: Some(reply),
+                headers: None,
+            })
+            .await?;
+        Ok(())
+    }
+
+    pub async fn publish_with_reply_and_headers(
+        &mut self,
+        subject: String,
+        reply: String,
+        headers: HeaderMap,
+        payload: Bytes,
+    ) -> Result<(), Error> {
+        self.sender
+            .send(Command::Publish {
+                subject,
+                payload,
+                respond: Some(reply),
+                headers: Some(headers),
             })
             .await?;
         Ok(())
@@ -905,6 +778,26 @@ impl Client {
         }
     }
 
+    pub async fn request_with_headers(
+        &mut self,
+        subject: String,
+        headers: HeaderMap,
+        payload: Bytes,
+    ) -> Result<Message, Error> {
+        let inbox = self.new_inbox();
+        let mut sub = self.subscribe(inbox.clone()).await?;
+        self.publish_with_reply_and_headers(subject, inbox, headers, payload)
+            .await?;
+        self.flush().await?;
+        match sub.next().await {
+            Some(message) => Ok(message),
+            None => Err(Box::new(io::Error::new(
+                ErrorKind::BrokenPipe,
+                "did not receive any message",
+            ))),
+        }
+    }
+
     /// Create a new globally unique inbox which can be used for replies.
     ///
     /// # Examples
@@ -941,23 +834,15 @@ impl Client {
         subject: String,
         queue_group: Option<String>,
     ) -> Result<Subscriber, io::Error> {
+        let sid = self.next_subscription_id.fetch_add(1, Ordering::Relaxed);
         let (sender, receiver) = mpsc::channel(16);
 
-        // Aiming to make this the only lock (aside from internal locks in channels).
-        let mut context = self.subscription_context.lock().await;
-        let sid = context.insert(Subscription {
-            sender,
-            delivered: 0,
-            max: None,
-            subject: subject.to_owned(),
-            queue_group: queue_group.to_owned(),
-        });
-
         self.sender
             .send(Command::Subscribe {
                 sid,
                 subject,
                 queue_group,
+                sender,
             })
             .await
             .unwrap();
@@ -992,21 +877,30 @@ pub async fn connect_with_options<A: ToServerAddrs>(
     addrs: A,
     options: ConnectOptions,
 ) -> Result<Client, io::Error> {
-    let mut connector = Connector {
-        server_addrs: addrs.to_server_addrs()?.into_iter().collect(),
-        options: options.clone(),
+    let tls_required = options.tls_required;
+    let ping_interval = options.ping_interval;
+    let flush_interval = options.flush_interval;
+
+    let tls_options = TlsOptions {
+        tls_required: options.tls_required,
+        certificates: options.certificates,
+        client_key: options.client_key,
+        client_cert: options.client_cert,
+        tls_client_config: options.tls_client_config,
     };
 
+    let mut connector = Connector::new(addrs, tls_options)?;
     let (_, connection) = connector.try_connect().await?;
-    let subscription_context = Arc::new(Mutex::new(SubscriptionContext::new()));
-    let mut connection_handler =
-        ConnectionHandler::new(connection, connector, subscription_context.clone());
+    let (events_tx, mut events_rx) = mpsc::channel(128);
+
+    let mut connection_handler = ConnectionHandler::new(connection, connector, events_tx);
 
     // TODO make channel size configurable
     let (sender, receiver) = mpsc::channel(128);
-    let client = Client::new(sender.clone(), subscription_context);
-    let connect_info = ConnectInfo {
-        tls_required: options.tls_required,
+
+    let client = Client::new(sender.clone());
+    let mut connect_info = ConnectInfo {
+        tls_required,
         // FIXME(tp): have optional name
         name: Some("beta-rust-client".to_string()),
         pedantic: false,
@@ -1024,11 +918,29 @@ pub async fn connect_with_options<A: ToServerAddrs>(
         headers: true,
         no_responders: true,
     };
-    client
-        .sender
-        .send(Command::Connect(connect_info))
-        .await
-        .map_err(|_| io::Error::new(io::ErrorKind::Other, "failed to send connect"))?;
+    match options.auth {
+        Authorization::None => {}
+        Authorization::Token(token) => connect_info.auth_token = Some(token),
+        Authorization::UserAndPassword(user, pass) => {
+            connect_info.user = Some(user);
+            connect_info.pass = Some(pass);
+        }
+    }
+    connection_handler
+        .connection
+        .write_op(ClientOp::Connect(connect_info))
+        .await?;
+    connection_handler.connection.flush().await?;
+    match connection_handler.connection.read_op().await? {
+        Some(ServerOp::Error(err)) => {
+            return Err(io::Error::new(ErrorKind::InvalidInput, err.to_string()));
+        }
+        Some(op) => {
+            connection_handler.handle_server_op(op).await?;
+        }
+        None => return Err(io::Error::new(ErrorKind::BrokenPipe, "connection aborted")),
+    }
+
     client
         .sender
         .send(Command::Ping)
@@ -1039,7 +951,7 @@ pub async fn connect_with_options<A: ToServerAddrs>(
         let sender = sender.clone();
         async move {
             loop {
-                tokio::time::sleep(options.ping_interval).await;
+                tokio::time::sleep(ping_interval).await;
                 match sender.send(Command::Ping).await {
                     Ok(()) => {}
                     Err(_) => return,
@@ -1050,7 +962,7 @@ pub async fn connect_with_options<A: ToServerAddrs>(
 
     tokio::spawn(async move {
         loop {
-            tokio::time::sleep(options.flush_interval).await;
+            tokio::time::sleep(flush_interval).await;
             match sender.send(Command::TryFlush).await {
                 Ok(()) => {}
                 Err(_) => return,
@@ -1058,11 +970,27 @@ pub async fn connect_with_options<A: ToServerAddrs>(
         }
     });
 
+    task::spawn(async move {
+        while let Some(event) = events_rx.recv().await {
+            match event {
+                ServerEvent::Reconnect => options.reconnect_callback.call().await,
+                ServerEvent::Disconnect => options.disconnect_callback.call().await,
+                ServerEvent::Error(error) => options.error_callback.call(error).await,
+            }
+        }
+    });
+
     task::spawn(async move { connection_handler.process(receiver).await });
 
     Ok(client)
 }
 
+pub(crate) enum ServerEvent {
+    Reconnect,
+    Disconnect,
+    Error(ServerError),
+}
+
 /// Connects to NATS with default config.
 ///
 /// Returns clonable [Client].
@@ -1087,6 +1015,7 @@ pub struct Message {
     pub subject: String,
     pub reply: Option<String>,
     pub payload: Bytes,
+    pub headers: Option<HeaderMap>,
 }
 
 /// Retrieves messages from given `subscription` created by [Client::subscribe].
@@ -1103,19 +1032,19 @@ pub struct Message {
 /// # }
 /// ```
 pub struct Subscriber {
-    uid: u64,
+    sid: u64,
     receiver: mpsc::Receiver<Message>,
     sender: mpsc::Sender<Command>,
 }
 
 impl Subscriber {
     fn new(
-        uid: u64,
+        sid: u64,
         sender: mpsc::Sender<Command>,
         receiver: mpsc::Receiver<Message>,
     ) -> Subscriber {
         Subscriber {
-            uid,
+            sid,
             sender,
             receiver,
         }
@@ -1137,7 +1066,7 @@ impl Subscriber {
     pub async fn unsubscribe(&mut self) -> io::Result<()> {
         self.sender
             .send(Command::Unsubscribe {
-                uid: self.uid,
+                sid: self.sid,
                 max: None,
             })
             .await
@@ -1174,7 +1103,7 @@ impl Subscriber {
     pub async fn unsubscribe_after(&mut self, unsub_after: u64) -> io::Result<()> {
         self.sender
             .send(Command::Unsubscribe {
-                uid: self.uid,
+                sid: self.sid,
                 max: Some(unsub_after),
             })
             .await
@@ -1188,10 +1117,10 @@ impl Drop for Subscriber {
         self.receiver.close();
         tokio::spawn({
             let sender = self.sender.clone();
-            let id = self.uid;
+            let sid = self.sid;
             async move {
                 sender
-                    .send(Command::Unsubscribe { uid: id, max: None })
+                    .send(Command::Unsubscribe { sid, max: None })
                     .await
                     .ok();
             }
@@ -1207,6 +1136,30 @@ impl Stream for Subscriber {
     }
 }
 
+#[derive(Clone, Debug, Eq, PartialEq)]
+pub enum ServerError {
+    AuthorizationViolation,
+    Other(String),
+}
+
+impl ServerError {
+    fn new(error: String) -> ServerError {
+        match error.to_lowercase().as_str() {
+            "authorization violation" => ServerError::AuthorizationViolation,
+            other => ServerError::Other(other.to_string()),
+        }
+    }
+}
+
+impl std::fmt::Display for ServerError {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        match self {
+            Self::AuthorizationViolation => write!(f, "nats: authorization violation"),
+            Self::Other(error) => write!(f, "nats: {}", error),
+        }
+    }
+}
+
 /// Info to construct a CONNECT message.
 #[derive(Clone, Debug, Serialize)]
 #[doc(hidden)]
@@ -1417,3 +1370,15 @@ impl<T: ToServerAddrs + ?Sized> ToServerAddrs for &T {
         (**self).to_server_addrs()
     }
 }
+
+#[derive(Clone)]
+pub(crate) enum Authorization {
+    /// No authentication.
+    None,
+
+    /// Authenticate using a token.
+    Token(String),
+
+    /// Authenticate using a username and password.
+    UserAndPassword(String, String),
+}
