@@ -15,6 +15,8 @@ mod nats_server;
 
 mod client {
 
+    use std::{collections::HashMap, time::Duration};
+
     use super::nats_server;
     use bytes::Bytes;
     use futures_util::{future::join_all, StreamExt};
@@ -121,6 +123,27 @@ mod client {
         assert_eq!(i, 10);
     }
 
+    #[tokio::test]
+    async fn publish_with_headers() {
+        let server = nats_server::run_basic_server();
+        let mut client = async_nats::connect(server.client_url()).await.unwrap();
+
+        let mut subscriber = client.subscribe("test".into()).await.unwrap();
+
+        let mut headers = async_nats::HeaderMap::new();
+        headers.append("X-Test", b"Test".as_ref().try_into().unwrap());
+
+        client
+            .publish_with_headers("test".into(), headers.clone(), b"".as_ref().into())
+            .await
+            .unwrap();
+
+        client.flush().await.unwrap();
+
+        let message = subscriber.next().await.unwrap();
+        assert_eq!(message.headers.unwrap(), headers);
+    }
+
     #[tokio::test]
     async fn publish_request() {
         let server = nats_server::run_basic_server();
@@ -204,7 +227,6 @@ mod client {
             client.publish("test".into(), "data".into()).await.unwrap();
         }
 
-        client.flush().await.unwrap();
         sub.unsubscribe_after(3).await.unwrap();
         client.publish("test".into(), "data".into()).await.unwrap();
         client.flush().await.unwrap();
@@ -223,7 +245,6 @@ mod client {
 
         client.publish("test".into(), "data".into()).await.unwrap();
         client.publish("test".into(), "data".into()).await.unwrap();
-        client.flush().await.unwrap();
 
         sub.unsubscribe_after(1).await.unwrap();
         client.flush().await.unwrap();
@@ -247,23 +268,109 @@ mod client {
             nats_server::run_basic_server(),
         ];
 
-        let mut client = async_nats::connect(
-            servers
-                .iter()
-                .map(|server| server.client_url().parse::<ServerAddr>().unwrap())
-                .collect::<Vec<ServerAddr>>()
-                .as_slice(),
-        )
-        .await
-        .unwrap();
+        let mut client = async_nats::ConnectOptions::new()
+            .connect(
+                servers
+                    .iter()
+                    .map(|server| server.client_url().parse::<ServerAddr>().unwrap())
+                    .collect::<Vec<ServerAddr>>()
+                    .as_slice(),
+            )
+            .await
+            .unwrap();
 
         let mut subscriber = client.subscribe("test".into()).await.unwrap();
         while !servers.is_empty() {
             client.publish("test".into(), "data".into()).await.unwrap();
+            client.flush().await.unwrap();
             assert!(subscriber.next().await.is_some());
 
             drop(servers.remove(0));
             tokio::time::sleep(std::time::Duration::from_secs(3)).await;
         }
     }
+
+    #[tokio::test]
+    async fn token_auth() {
+        let server = nats_server::run_server("tests/configs/token.conf");
+        let mut nc = async_nats::ConnectOptions::with_token("s3cr3t".into())
+            .connect(server.client_url())
+            .await
+            .unwrap();
+
+        let mut sub = nc.subscribe("test".into()).await.unwrap();
+        nc.publish("test".into(), "test".into()).await.unwrap();
+        nc.flush().await.unwrap();
+        assert!(sub.next().await.is_some());
+    }
+
+    #[tokio::test]
+    async fn user_pass_auth() {
+        let server = nats_server::run_server("tests/configs/user_pass.conf");
+        let mut nc =
+            async_nats::ConnectOptions::with_user_and_password("derek".into(), "s3cr3t".into())
+                .connect(server.client_url())
+                .await
+                .unwrap();
+
+        let mut sub = nc.subscribe("test".into()).await.unwrap();
+        nc.publish("test".into(), "test".into()).await.unwrap();
+        nc.flush().await.unwrap();
+        assert!(sub.next().await.is_some());
+    }
+
+    #[tokio::test]
+    async fn user_pass_auth_wrong_pass() {
+        let server = nats_server::run_server("tests/configs/user_pass.conf");
+        async_nats::ConnectOptions::with_user_and_password("derek".into(), "bad_password".into())
+            .connect(server.client_url())
+            .await
+            .unwrap_err();
+    }
+
+    #[tokio::test]
+    async fn connection_callbacks() {
+        let server = nats_server::run_basic_server();
+        let port = server.client_port().to_string();
+
+        let (tx, mut rx) = tokio::sync::mpsc::channel(128);
+        let (dc_tx, mut dc_rx) = tokio::sync::mpsc::channel(128);
+        let mut nc = async_nats::ConnectOptions::new()
+            .reconnect_callback(move || {
+                let tx = tx.clone();
+                async move {
+                    println!("reconnection callback fired");
+                    tx.send(()).await.unwrap();
+                }
+            })
+            .disconnect_callback(move || {
+                let dc_tx = dc_tx.clone();
+                async move {
+                    println!("disconnect callback fired");
+                    dc_tx.send(()).await.unwrap();
+                }
+            })
+            .connect(server.client_url())
+            .await
+            .unwrap();
+        println!("conncted");
+        nc.subscribe("test".to_string()).await.unwrap();
+        nc.flush().await.unwrap();
+
+        println!("dropped server {:?}", server.client_url());
+        drop(server);
+        tokio::time::sleep(Duration::from_secs(3)).await;
+
+        let _server = nats_server::run_server_with_port("", Some(port.as_str()));
+
+        tokio::time::timeout(Duration::from_secs(15), dc_rx.recv())
+            .await
+            .unwrap()
+            .unwrap();
+
+        tokio::time::timeout(Duration::from_secs(15), rx.recv())
+            .await
+            .unwrap()
+            .unwrap();
+    }
 }
